@@ -26,6 +26,9 @@
 
 #include "gstmpeg2dec.h"
 
+/* 16byte-aligns a buffer for libmpeg2 */
+#define ALIGN_16(p) ((void *)(((uintptr_t)(p) + 15) & ~((uintptr_t)15)))
+
 /* mpeg2dec changed a struct name after 0.3.1, here's a workaround */
 /* mpeg2dec also only defined MPEG2_RELEASE after 0.3.1
    #if MPEG2_RELEASE < MPEG2_VERSION(0,3,2)
@@ -42,18 +45,6 @@ typedef gint mpeg2_state_t;
 GST_DEBUG_CATEGORY_STATIC (mpeg2dec_debug);
 #define GST_CAT_DEFAULT (mpeg2dec_debug)
 
-/* table with framerates expressed as fractions */
-static const gint fpss[][2] = { {24000, 1001},
-{24, 1}, {25, 1}, {30000, 1001},
-{30, 1}, {50, 1}, {60000, 1001},
-{60, 1}, {0, 1}
-};
-
-/* frame periods */
-static const guint frame_periods[] = {
-  1126125, 1125000, 1080000, 900900, 900000, 540000, 450450, 450000, 0
-};
-
 /* elementfactory information */
 static const GstElementDetails gst_mpeg2dec_details =
 GST_ELEMENT_DETAILS ("mpeg1 and mpeg2 video decoder",
@@ -61,11 +52,12 @@ GST_ELEMENT_DETAILS ("mpeg1 and mpeg2 video decoder",
     "Uses libmpeg2 to decode MPEG video streams",
     "Wim Taymans <wim.taymans@chello.be>");
 
-/* error out after receiving MAX_ERROR_COUNT STATE_INVALID return value
- * from mpeg2_parse. -1 means never error out
+/* Send a warning message about decoding errors after receiving this many
+ * STATE_INVALID return values from mpeg2_parse. -1 means never.
  */
-#define MAX_ERROR_COUNT (5)
+#define WARN_THRESHOLD (5)
 
+//#define enable_user_data
 #ifdef enable_user_data
 static GstStaticPadTemplate user_data_template_factory =
 GST_STATIC_PAD_TEMPLATE ("user_data",
@@ -87,10 +79,10 @@ GST_STATIC_PAD_TEMPLATE ("src",
     GST_PAD_SRC,
     GST_PAD_ALWAYS,
     GST_STATIC_CAPS ("video/x-raw-yuv, "
-        "format = (fourcc) { YV12, I420, Y42B }, "
+        "format = (fourcc) { YV12, I420, Y42B, Y444 }, "
         "width = (int) [ 16, 4096 ], "
         "height = (int) [ 16, 4096 ], "
-        "framerate = (fraction) { 24000/1001, 24/1, 25/1, 30000/1001, 30/1, 50/1, 60000/1001, 60/1 }")
+        "framerate = (fraction) [ 0/1, 2147483647/1 ]")
     );
 
 static void gst_mpeg2dec_base_init (gpointer g_class);
@@ -100,12 +92,10 @@ static void gst_mpeg2dec_init (GstMpeg2dec * mpeg2dec);
 static void gst_mpeg2dec_finalize (GObject * object);
 static void gst_mpeg2dec_reset (GstMpeg2dec * mpeg2dec);
 
-static void gst_mpeg2dec_set_property (GObject * object, guint prop_id,
-    const GValue * value, GParamSpec * pspec);
-static void gst_mpeg2dec_get_property (GObject * object, guint prop_id,
-    GValue * value, GParamSpec * pspec);
+#ifndef GST_DISABLE_INDEX
 static void gst_mpeg2dec_set_index (GstElement * element, GstIndex * index);
 static GstIndex *gst_mpeg2dec_get_index (GstElement * element);
+#endif
 
 static gboolean gst_mpeg2dec_src_event (GstPad * pad, GstEvent * event);
 static const GstQueryType *gst_mpeg2dec_get_src_query_types (GstPad * pad);
@@ -121,6 +111,7 @@ static GstStateChangeReturn gst_mpeg2dec_change_state (GstElement * element,
     GstStateChange transition);
 
 static gboolean gst_mpeg2dec_sink_event (GstPad * pad, GstEvent * event);
+static gboolean gst_mpeg2dec_setcaps (GstPad * pad, GstCaps * caps);
 static GstFlowReturn gst_mpeg2dec_chain (GstPad * pad, GstBuffer * buf);
 
 static void clear_buffers (GstMpeg2dec * mpeg2dec);
@@ -197,13 +188,13 @@ gst_mpeg2dec_class_init (GstMpeg2decClass * klass)
 
   parent_class = g_type_class_peek_parent (klass);
 
-  gobject_class->set_property = gst_mpeg2dec_set_property;
-  gobject_class->get_property = gst_mpeg2dec_get_property;
   gobject_class->finalize = gst_mpeg2dec_finalize;
 
   gstelement_class->change_state = gst_mpeg2dec_change_state;
+#ifndef GST_DISABLE_INDEX
   gstelement_class->set_index = gst_mpeg2dec_set_index;
   gstelement_class->get_index = gst_mpeg2dec_get_index;
+#endif
 }
 
 static void
@@ -211,8 +202,7 @@ gst_mpeg2dec_init (GstMpeg2dec * mpeg2dec)
 {
   /* create the sink and src pads */
   mpeg2dec->sinkpad =
-      gst_pad_new_from_template (gst_static_pad_template_get
-      (&sink_template_factory), "sink");
+      gst_pad_new_from_static_template (&sink_template_factory, "sink");
   gst_pad_set_chain_function (mpeg2dec->sinkpad,
       GST_DEBUG_FUNCPTR (gst_mpeg2dec_chain));
 #if 0
@@ -221,11 +211,12 @@ gst_mpeg2dec_init (GstMpeg2dec * mpeg2dec)
 #endif
   gst_pad_set_event_function (mpeg2dec->sinkpad,
       GST_DEBUG_FUNCPTR (gst_mpeg2dec_sink_event));
+  gst_pad_set_setcaps_function (mpeg2dec->sinkpad,
+      GST_DEBUG_FUNCPTR (gst_mpeg2dec_setcaps));
   gst_element_add_pad (GST_ELEMENT (mpeg2dec), mpeg2dec->sinkpad);
 
   mpeg2dec->srcpad =
-      gst_pad_new_from_template (gst_static_pad_template_get
-      (&src_template_factory), "src");
+      gst_pad_new_from_static_template (&src_template_factory, "src");
   gst_pad_set_event_function (mpeg2dec->srcpad,
       GST_DEBUG_FUNCPTR (gst_mpeg2dec_src_event));
   gst_pad_set_query_type_function (mpeg2dec->srcpad,
@@ -237,12 +228,13 @@ gst_mpeg2dec_init (GstMpeg2dec * mpeg2dec)
 
 #ifdef enable_user_data
   mpeg2dec->userdatapad =
-      gst_pad_new_from_template (gst_static_pad_template_get
-      (&user_data_template_factory), "user_data");
+      gst_pad_new_from_static_template (&user_data_template_factory,
+      "user_data");
   gst_element_add_pad (GST_ELEMENT (mpeg2dec), mpeg2dec->userdatapad);
 #endif
 
   mpeg2dec->error_count = 0;
+  mpeg2dec->can_allocate_aligned = TRUE;
 
   /* initialize the mpeg2dec acceleration */
 }
@@ -258,8 +250,8 @@ gst_mpeg2dec_finalize (GObject * object)
     mpeg2dec->decoder = NULL;
   }
   clear_buffers (mpeg2dec);
-  g_free (mpeg2dec->dummybuf[0]);
-  mpeg2dec->dummybuf[0] = NULL;
+  g_free (mpeg2dec->dummybuf[3]);
+  mpeg2dec->dummybuf[3] = NULL;
 
   G_OBJECT_CLASS (parent_class)->finalize (object);
 }
@@ -278,6 +270,7 @@ gst_mpeg2dec_reset (GstMpeg2dec * mpeg2dec)
   mpeg2dec->next_time = -1;
   mpeg2dec->offset = 0;
   mpeg2dec->error_count = 0;
+  mpeg2dec->can_allocate_aligned = TRUE;
   mpeg2_reset (mpeg2dec->decoder, 1);
 }
 
@@ -290,6 +283,7 @@ gst_mpeg2dec_qos_reset (GstMpeg2dec * mpeg2dec)
   GST_OBJECT_UNLOCK (mpeg2dec);
 }
 
+#ifndef GST_DISABLE_INDEX
 static void
 gst_mpeg2dec_set_index (GstElement * element, GstIndex * index)
 {
@@ -307,6 +301,7 @@ gst_mpeg2dec_get_index (GstElement * element)
 
   return mpeg2dec->index;
 }
+#endif
 
 /* see gst-plugins/gst/games/gstvideoimage.c, paint_setup_I420() */
 #define I420_Y_ROWSTRIDE(width) (GST_ROUND_UP_4(width))
@@ -403,9 +398,9 @@ crop_copy_i422_buffer (GstMpeg2dec * mpeg2dec, GstBuffer * input)
 static gboolean
 crop_buffer (GstMpeg2dec * mpeg2dec, GstBuffer ** buf)
 {
-  gboolean result = FALSE;
+  gboolean result = TRUE;
   GstBuffer *input = *buf;
-  GstBuffer *outbuf = input;
+  GstBuffer *outbuf;
 
   /*We crop only if the target region is smaller than the input one */
   if ((mpeg2dec->decoded_width > mpeg2dec->width) ||
@@ -427,7 +422,7 @@ crop_buffer (GstMpeg2dec * mpeg2dec, GstBuffer ** buf)
       }
 
       gst_buffer_set_caps (outbuf, GST_PAD_CAPS (mpeg2dec->srcpad));
-      gst_buffer_stamp (outbuf, input);
+      gst_buffer_copy_metadata (outbuf, input, GST_BUFFER_COPY_TIMESTAMPS);
       gst_buffer_unref (input);
 
       *buf = outbuf;
@@ -442,22 +437,41 @@ static GstFlowReturn
 gst_mpeg2dec_alloc_sized_buf (GstMpeg2dec * mpeg2dec, guint size,
     GstBuffer ** obuf)
 {
-  GstFlowReturn ret;
+  if (mpeg2dec->can_allocate_aligned
+      && mpeg2dec->decoded_width == mpeg2dec->width
+      && mpeg2dec->decoded_height == mpeg2dec->height) {
+    GstFlowReturn ret;
 
-  if (mpeg2dec->decoded_width == mpeg2dec->width &&
-      mpeg2dec->decoded_height == mpeg2dec->height) {
     ret = gst_pad_alloc_buffer_and_set_caps (mpeg2dec->srcpad,
         GST_BUFFER_OFFSET_NONE, size, GST_PAD_CAPS (mpeg2dec->srcpad), obuf);
-  } else {
-    /* can't use gst_pad_alloc_buffer() here because the output buffer will
-     * be cropped and basetransform-based elements will complain about
-     * the wrong unit size when not operating in passthrough mode */
-    *obuf = gst_buffer_new_and_alloc (size);
-    gst_buffer_set_caps (*obuf, GST_PAD_CAPS (mpeg2dec->srcpad));
-    ret = GST_FLOW_OK;
+    if (ret != GST_FLOW_OK) {
+      return ret;
+    }
+
+    /* libmpeg2 needs 16 byte aligned buffers... test for this here
+     * and if it fails only a single time create our own buffers from
+     * there on below that are correctly aligned */
+    if (((uintptr_t) GST_BUFFER_DATA (*obuf)) % 16 == 0) {
+      GST_LOG_OBJECT (mpeg2dec, "return 16 byte aligned buffer");
+      return ret;
+    }
+
+    GST_DEBUG_OBJECT (mpeg2dec,
+        "can't get 16 byte aligned buffers, creating our own ones");
+    gst_buffer_unref (*obuf);
+    mpeg2dec->can_allocate_aligned = FALSE;
   }
 
-  return ret;
+  /* can't use gst_pad_alloc_buffer() here because the output buffer will
+   * either be cropped later or be bigger than expected (for the alignment),
+   * and basetransform-based elements will complain about the wrong unit size
+   * when not operating in passthrough mode */
+  *obuf = gst_buffer_new_and_alloc (size + 15);
+  GST_BUFFER_DATA (*obuf) = (guint8 *) ALIGN_16 (GST_BUFFER_DATA (*obuf));
+  GST_BUFFER_SIZE (*obuf) = size;
+  gst_buffer_set_caps (*obuf, GST_PAD_CAPS (mpeg2dec->srcpad));
+
+  return GST_FLOW_OK;
 }
 
 static GstFlowReturn
@@ -528,8 +542,10 @@ gst_mpeg2dec_negotiate_format (GstMpeg2dec * mpeg2dec)
     mpeg2dec->v_offs =
         I420_V_OFFSET (mpeg2dec->decoded_width, mpeg2dec->decoded_height);
 
-  } else if (sequence->width == sequence->chroma_width ||
-      sequence->height == sequence->chroma_height) {
+  } else if ((sequence->width == sequence->chroma_width &&
+          sequence->height != sequence->chroma_height) ||
+      (sequence->width != sequence->chroma_width &&
+          sequence->height == sequence->chroma_height)) {
     gint halfsize;
 
     fourcc = GST_STR_FOURCC ("Y42B");
@@ -539,8 +555,15 @@ gst_mpeg2dec_negotiate_format (GstMpeg2dec * mpeg2dec)
     mpeg2dec->u_offs = halfsize;
     mpeg2dec->v_offs = halfsize + (halfsize / 2);
   } else {
-    g_warning ("mpeg2dec: 4:4:4 format not yet supported");
-    return (FALSE);
+    gint size;
+
+    size = mpeg2dec->decoded_width * mpeg2dec->decoded_height;
+
+    fourcc = GST_STR_FOURCC ("Y444");
+    mpeg2dec->format = MPEG2DEC_FORMAT_Y444;
+    mpeg2dec->size = size * 3;
+    mpeg2dec->u_offs = size;
+    mpeg2dec->v_offs = size * 2;
   }
 
   if (mpeg2dec->pixel_width == 0 || mpeg2dec->pixel_height == 0) {
@@ -578,7 +601,8 @@ gst_mpeg2dec_negotiate_format (GstMpeg2dec * mpeg2dec)
       "height", G_TYPE_INT, mpeg2dec->height,
       "pixel-aspect-ratio", GST_TYPE_FRACTION, mpeg2dec->pixel_width,
       mpeg2dec->pixel_height,
-      "framerate", GST_TYPE_FRACTION, mpeg2dec->fps_n, mpeg2dec->fps_d, NULL);
+      "framerate", GST_TYPE_FRACTION, mpeg2dec->fps_n, mpeg2dec->fps_d,
+      "interlaced", G_TYPE_BOOLEAN, mpeg2dec->interlaced, NULL);
 
   gst_pad_set_caps (mpeg2dec->srcpad, caps);
   gst_caps_unref (caps);
@@ -589,36 +613,55 @@ gst_mpeg2dec_negotiate_format (GstMpeg2dec * mpeg2dec)
 static void
 init_dummybuf (GstMpeg2dec * mpeg2dec)
 {
-  g_free (mpeg2dec->dummybuf[0]);
+  g_free (mpeg2dec->dummybuf[3]);
 
-  mpeg2dec->dummybuf[0] = g_malloc (mpeg2dec->size);
+  /* libmpeg2 needs 16 byte aligned buffers... care for this here */
+  mpeg2dec->dummybuf[3] = g_malloc0 (mpeg2dec->size + 15);
+  mpeg2dec->dummybuf[0] = ALIGN_16 (mpeg2dec->dummybuf[3]);
   mpeg2dec->dummybuf[1] = mpeg2dec->dummybuf[0] + mpeg2dec->u_offs;
-  mpeg2dec->dummybuf[2] = mpeg2dec->dummybuf[1] + mpeg2dec->v_offs;
+  mpeg2dec->dummybuf[2] = mpeg2dec->dummybuf[0] + mpeg2dec->v_offs;
 }
 
 static GstFlowReturn
 handle_sequence (GstMpeg2dec * mpeg2dec, const mpeg2_info_t * info)
 {
-  gint i;
   GstFlowReturn ret = GST_FLOW_OK;
 
+  if (info->sequence->frame_period == 0) {
+    GST_WARNING_OBJECT (mpeg2dec, "Frame period is 0!");
+    ret = GST_FLOW_ERROR;
+    goto done;
+  }
+
   mpeg2dec->width = info->sequence->picture_width;
   mpeg2dec->height = info->sequence->picture_height;
-  mpeg2dec->pixel_width = info->sequence->pixel_width;
-  mpeg2dec->pixel_height = info->sequence->pixel_height;
   mpeg2dec->decoded_width = info->sequence->width;
   mpeg2dec->decoded_height = info->sequence->height;
   mpeg2dec->total_frames = 0;
 
-  /* find framerate */
-  for (i = 0; i < 9; i++) {
-    if (info->sequence->frame_period == frame_periods[i]) {
-      mpeg2dec->fps_n = fpss[i][0];
-      mpeg2dec->fps_d = fpss[i][1];
-    }
+  /* don't take the sequence PAR if we already have one from the sink caps */
+  if (!mpeg2dec->have_par) {
+    mpeg2dec->pixel_width = info->sequence->pixel_width;
+    mpeg2dec->pixel_height = info->sequence->pixel_height;
   }
+
+  /* mpeg2 video can only be from 16x16 to 4096x4096. Everything
+   * else is a corrupted files */
+  if (mpeg2dec->width > 4096 || mpeg2dec->width < 16 ||
+      mpeg2dec->height > 4096 || mpeg2dec->height < 16) {
+    GST_ERROR_OBJECT (mpeg2dec, "Invalid frame dimensions: %d x %d",
+        mpeg2dec->width, mpeg2dec->height);
+    return GST_FLOW_ERROR;
+  }
+
+  /* set framerate */
+  mpeg2dec->fps_n = 27000000;
+  mpeg2dec->fps_d = info->sequence->frame_period;
   mpeg2dec->frame_period = info->sequence->frame_period * GST_USECOND / 27;
 
+  mpeg2dec->interlaced =
+      !(info->sequence->flags & SEQ_FLAG_PROGRESSIVE_SEQUENCE);
+
   GST_DEBUG_OBJECT (mpeg2dec,
       "sequence flags: %d, frame period: %d (%g), frame rate: %d/%d",
       info->sequence->flags, info->sequence->frame_period,
@@ -629,6 +672,13 @@ handle_sequence (GstMpeg2dec * mpeg2dec, const mpeg2_info_t * info)
   GST_DEBUG_OBJECT (mpeg2dec, "transfer chars: %d, matrix coef: %d",
       info->sequence->transfer_characteristics,
       info->sequence->matrix_coefficients);
+  GST_DEBUG_OBJECT (mpeg2dec,
+      "FLAGS: CONSTRAINED_PARAMETERS:%d, PROGRESSIVE_SEQUENCE:%d",
+      info->sequence->flags & SEQ_FLAG_CONSTRAINED_PARAMETERS,
+      info->sequence->flags & SEQ_FLAG_PROGRESSIVE_SEQUENCE);
+  GST_DEBUG_OBJECT (mpeg2dec, "FLAGS: LOW_DELAY:%d, COLOUR_DESCRIPTION:%d",
+      info->sequence->flags & SEQ_FLAG_LOW_DELAY,
+      info->sequence->flags & SEQ_FLAG_COLOUR_DESCRIPTION);
 
   if (!gst_mpeg2dec_negotiate_format (mpeg2dec))
     goto negotiate_failed;
@@ -674,6 +724,35 @@ clear_buffers (GstMpeg2dec * mpeg2dec)
   *bufpen = NULL;
 }
 
+static void
+clear_queued (GstMpeg2dec * mpeg2dec)
+{
+  g_list_foreach (mpeg2dec->queued, (GFunc) gst_mini_object_unref, NULL);
+  g_list_free (mpeg2dec->queued);
+  mpeg2dec->queued = NULL;
+}
+
+static GstFlowReturn
+flush_queued (GstMpeg2dec * mpeg2dec)
+{
+  GstFlowReturn res = GST_FLOW_OK;
+
+  while (mpeg2dec->queued) {
+    GstBuffer *buf = GST_BUFFER_CAST (mpeg2dec->queued->data);
+
+    GST_LOG_OBJECT (mpeg2dec, "pushing buffer %p, timestamp %"
+        GST_TIME_FORMAT ", duration %" GST_TIME_FORMAT, buf,
+        GST_TIME_ARGS (GST_BUFFER_TIMESTAMP (buf)),
+        GST_TIME_ARGS (GST_BUFFER_DURATION (buf)));
+
+    /* iterate ouput queue an push downstream */
+    res = gst_pad_push (mpeg2dec->srcpad, buf);
+
+    mpeg2dec->queued = g_list_delete_link (mpeg2dec->queued, mpeg2dec->queued);
+  }
+  return res;
+}
+
 static GstFlowReturn
 handle_picture (GstMpeg2dec * mpeg2dec, const mpeg2_info_t * info)
 {
@@ -699,6 +778,11 @@ handle_picture (GstMpeg2dec * mpeg2dec, const mpeg2_info_t * info)
   switch (type) {
     case PIC_FLAG_CODING_TYPE_I:
       mpeg2_skip (mpeg2dec->decoder, 0);
+      if (mpeg2dec->segment.rate < 0.0) {
+        /* negative rate, flush the queued pictures in reverse */
+        GST_DEBUG_OBJECT (mpeg2dec, "flushing queued buffers");
+        flush_queued (mpeg2dec);
+      }
     case PIC_FLAG_CODING_TYPE_P:
       bufpen = &mpeg2dec->ip_buffers[mpeg2dec->ip_bufpos];
       GST_DEBUG_OBJECT (mpeg2dec, "I/P unref %p, ref %p", *bufpen, outbuf);
@@ -735,6 +819,49 @@ no_buffer:
   }
 }
 
+/* try to clip the buffer to the segment boundaries */
+static gboolean
+clip_buffer (GstMpeg2dec * dec, GstBuffer * buf)
+{
+  gboolean res = TRUE;
+  GstClockTime in_ts, in_dur, stop;
+  gint64 cstart, cstop;
+
+  in_ts = GST_BUFFER_TIMESTAMP (buf);
+  in_dur = GST_BUFFER_DURATION (buf);
+
+  GST_LOG_OBJECT (dec,
+      "timestamp:%" GST_TIME_FORMAT " , duration:%" GST_TIME_FORMAT,
+      GST_TIME_ARGS (in_ts), GST_TIME_ARGS (in_dur));
+
+  /* can't clip without TIME segment */
+  if (dec->segment.format != GST_FORMAT_TIME)
+    goto beach;
+
+  /* we need a start time */
+  if (!GST_CLOCK_TIME_IS_VALID (in_ts))
+    goto beach;
+
+  /* generate valid stop, if duration unknown, we have unknown stop */
+  stop =
+      GST_CLOCK_TIME_IS_VALID (in_dur) ? (in_ts + in_dur) : GST_CLOCK_TIME_NONE;
+
+  /* now clip */
+  if (!(res = gst_segment_clip (&dec->segment, GST_FORMAT_TIME,
+              in_ts, stop, &cstart, &cstop)))
+    goto beach;
+
+  /* update timestamp and possibly duration if the clipped stop time is
+   * valid */
+  GST_BUFFER_TIMESTAMP (buf) = cstart;
+  if (GST_CLOCK_TIME_IS_VALID (cstop))
+    GST_BUFFER_DURATION (buf) = cstop - cstart;
+
+beach:
+  GST_LOG_OBJECT (dec, "%sdropping", (res ? "not " : ""));
+  return res;
+}
+
 static GstFlowReturn
 handle_slice (GstMpeg2dec * mpeg2dec, const mpeg2_info_t * info)
 {
@@ -774,12 +901,19 @@ handle_slice (GstMpeg2dec * mpeg2dec, const mpeg2_info_t * info)
   time = GST_CLOCK_TIME_NONE;
 
 #if MPEG2_RELEASE < MPEG2_VERSION(0,4,0)
-  if (picture->flags & PIC_FLAG_PTS)
+  if (picture->flags & PIC_FLAG_PTS) {
     time = MPEG_TIME_TO_GST_TIME (picture->pts);
+    GST_DEBUG_OBJECT (mpeg2dec, "picture pts %" G_GUINT64_FORMAT
+        ", time %" GST_TIME_FORMAT, picture->pts, GST_TIME_ARGS (time));
+  }
 #else
-  if (picture->flags & PIC_FLAG_TAGS)
-    time = MPEG_TIME_TO_GST_TIME ((GstClockTime) (picture->
-            tag2) << 32 | picture->tag);
+  if (picture->flags & PIC_FLAG_TAGS) {
+    guint64 pts = (((guint64) picture->tag2) << 32) | picture->tag;
+
+    time = MPEG_TIME_TO_GST_TIME (pts);
+    GST_DEBUG_OBJECT (mpeg2dec, "picture tags %" G_GUINT64_FORMAT
+        ", time %" GST_TIME_FORMAT, pts, GST_TIME_ARGS (time));
+  }
 #endif
 
   if (time == GST_CLOCK_TIME_NONE) {
@@ -804,20 +938,38 @@ handle_slice (GstMpeg2dec * mpeg2dec, const mpeg2_info_t * info)
   }
   mpeg2dec->next_time += GST_BUFFER_DURATION (outbuf);
 
+  if (picture->flags & PIC_FLAG_TOP_FIELD_FIRST)
+    GST_BUFFER_FLAG_SET (outbuf, GST_VIDEO_BUFFER_TFF);
+
+#if MPEG2_RELEASE >= MPEG2_VERSION(0,5,0)
+  /* repeat field introduced in 0.5.0 */
+  if (picture->flags & PIC_FLAG_REPEAT_FIRST_FIELD)
+    GST_BUFFER_FLAG_SET (outbuf, GST_VIDEO_BUFFER_RFF);
+#endif
+
   GST_DEBUG_OBJECT (mpeg2dec,
-      "picture: %s %s fields:%d off:%" G_GINT64_FORMAT " ts:%"
+      "picture: %s %s %s %s %s fields:%d off:%" G_GINT64_FORMAT " ts:%"
       GST_TIME_FORMAT,
-      (picture->flags & PIC_FLAG_TOP_FIELD_FIRST ? "tff " : "    "),
       (picture->flags & PIC_FLAG_PROGRESSIVE_FRAME ? "prog" : "    "),
+      (picture->flags & PIC_FLAG_TOP_FIELD_FIRST ? "tff" : "   "),
+#if MPEG2_RELEASE >= MPEG2_VERSION(0,5,0)
+      (picture->flags & PIC_FLAG_REPEAT_FIRST_FIELD ? "rff" : "   "),
+#else
+      "unknown rff",
+#endif
+      (picture->flags & PIC_FLAG_SKIP ? "skip" : "    "),
+      (picture->flags & PIC_FLAG_COMPOSITE_DISPLAY ? "composite" : "         "),
       picture->nb_fields, GST_BUFFER_OFFSET (outbuf),
       GST_TIME_ARGS (GST_BUFFER_TIMESTAMP (outbuf)));
 
+#ifndef GST_DISABLE_INDEX
   if (mpeg2dec->index) {
     gst_index_add_association (mpeg2dec->index, mpeg2dec->index_id,
         (key_frame ? GST_ASSOCIATION_FLAG_KEY_UNIT : 0),
         GST_FORMAT_BYTES, GST_BUFFER_OFFSET (outbuf),
         GST_FORMAT_TIME, GST_BUFFER_TIMESTAMP (outbuf), 0);
   }
+#endif
 
   if (picture->flags & PIC_FLAG_SKIP)
     goto skip;
@@ -825,6 +977,10 @@ handle_slice (GstMpeg2dec * mpeg2dec, const mpeg2_info_t * info)
   if (mpeg2dec->discont_state != MPEG2DEC_DISC_NONE)
     goto drop;
 
+  /* check for clipping */
+  if (!clip_buffer (mpeg2dec, outbuf))
+    goto clipped;
+
   if (GST_CLOCK_TIME_IS_VALID (time)) {
     gboolean need_skip;
     GstClockTime qostime;
@@ -846,12 +1002,6 @@ handle_slice (GstMpeg2dec * mpeg2dec, const mpeg2_info_t * info)
       goto dropping_qos;
   }
 
-  GST_LOG_OBJECT (mpeg2dec, "pushing buffer %p, timestamp %"
-      GST_TIME_FORMAT ", duration %" GST_TIME_FORMAT,
-      outbuf,
-      GST_TIME_ARGS (GST_BUFFER_TIMESTAMP (outbuf)),
-      GST_TIME_ARGS (GST_BUFFER_DURATION (outbuf)));
-
   /* ref before pushing it out, so we still have the ref in our
    * array of buffers */
   gst_buffer_ref (outbuf);
@@ -859,8 +1009,25 @@ handle_slice (GstMpeg2dec * mpeg2dec, const mpeg2_info_t * info)
   /* do cropping if needed */
   crop_buffer (mpeg2dec, &outbuf);
 
-  ret = gst_pad_push (mpeg2dec->srcpad, outbuf);
-  GST_DEBUG_OBJECT (mpeg2dec, "pushed with result %s", gst_flow_get_name (ret));
+  if (mpeg2dec->segment.rate >= 0.0) {
+    /* forward: push right away */
+    GST_LOG_OBJECT (mpeg2dec, "pushing buffer %p, timestamp %"
+        GST_TIME_FORMAT ", duration %" GST_TIME_FORMAT,
+        outbuf,
+        GST_TIME_ARGS (GST_BUFFER_TIMESTAMP (outbuf)),
+        GST_TIME_ARGS (GST_BUFFER_DURATION (outbuf)));
+    GST_LOG_OBJECT (mpeg2dec, "... with flags %x", GST_BUFFER_FLAGS (outbuf));
+
+    ret = gst_pad_push (mpeg2dec->srcpad, outbuf);
+    GST_DEBUG_OBJECT (mpeg2dec, "pushed with result %s",
+        gst_flow_get_name (ret));
+  } else {
+    /* reverse: queue, we'll push in reverse when we receive the next (previous)
+     * keyframe. */
+    GST_DEBUG_OBJECT (mpeg2dec, "queued frame");
+    mpeg2dec->queued = g_list_prepend (mpeg2dec->queued, outbuf);
+    ret = GST_FLOW_OK;
+  }
 
   return ret;
 
@@ -881,6 +1048,11 @@ drop:
         mpeg2dec->discont_state);
     return GST_FLOW_OK;
   }
+clipped:
+  {
+    GST_DEBUG_OBJECT (mpeg2dec, "dropping buffer, clipped");
+    return GST_FLOW_OK;
+  }
 dropping_qos:
   {
     GST_DEBUG_OBJECT (mpeg2dec, "dropping buffer because of QoS");
@@ -936,6 +1108,15 @@ gst_mpeg2dec_chain (GstPad * pad, GstBuffer * buf)
   data = GST_BUFFER_DATA (buf);
   pts = GST_BUFFER_TIMESTAMP (buf);
 
+  if (GST_BUFFER_IS_DISCONT (buf)) {
+    GST_LOG_OBJECT (mpeg2dec, "DISCONT, reset decoder");
+    /* when we receive a discont, reset our state as to not create too much
+     * distortion in the picture due to missing packets */
+    mpeg2_reset (mpeg2dec->decoder, 0);
+    mpeg2_skip (mpeg2dec->decoder, 1);
+    mpeg2dec->discont_state = MPEG2DEC_DISC_NEW_PICTURE;
+  }
+
   GST_LOG_OBJECT (mpeg2dec, "received buffer, timestamp %"
       GST_TIME_FORMAT ", duration %" GST_TIME_FORMAT,
       GST_TIME_ARGS (GST_BUFFER_TIMESTAMP (buf)),
@@ -973,8 +1154,32 @@ gst_mpeg2dec_chain (GstPad * pad, GstBuffer * buf)
     GST_DEBUG_OBJECT (mpeg2dec, "parse state %d", state);
 
     switch (state) {
+#if MPEG2_RELEASE >= MPEG2_VERSION (0, 5, 0)
+      case STATE_SEQUENCE_MODIFIED:
+        GST_DEBUG_OBJECT (mpeg2dec, "sequence modified");
+        /* fall through */
+#endif
       case STATE_SEQUENCE:
         ret = handle_sequence (mpeg2dec, info);
+        /* if there is an error handling the sequence
+         * reset the decoder, maybe something more elegant
+         * could be done.
+         */
+        if (ret == GST_FLOW_ERROR) {
+          mpeg2dec->error_count++;
+          GST_WARNING_OBJECT (mpeg2dec, "Decoding error #%d",
+              mpeg2dec->error_count);
+          if (mpeg2dec->error_count >= WARN_THRESHOLD && WARN_THRESHOLD > 0) {
+            GST_ELEMENT_WARNING (mpeg2dec, STREAM, DECODE,
+                ("%d consecutive decoding errors", mpeg2dec->error_count),
+                (NULL));
+          }
+          mpeg2_reset (mpeg2dec->decoder, 0);
+          mpeg2_skip (mpeg2dec->decoder, 1);
+          mpeg2dec->discont_state = MPEG2DEC_DISC_NEW_PICTURE;
+
+          goto exit;
+        }
         break;
       case STATE_SEQUENCE_REPEATED:
         GST_DEBUG_OBJECT (mpeg2dec, "sequence repeated");
@@ -1007,12 +1212,15 @@ gst_mpeg2dec_chain (GstPad * pad, GstBuffer * buf)
         break;
         /* error */
       case STATE_INVALID:
+        /* FIXME: at some point we should probably send newsegment events to
+         * let downstream know that parts of the stream are missing */
         mpeg2dec->error_count++;
         GST_WARNING_OBJECT (mpeg2dec, "Decoding error #%d",
             mpeg2dec->error_count);
-        if (mpeg2dec->error_count >= MAX_ERROR_COUNT && MAX_ERROR_COUNT > 0) {
-          GST_WARNING_OBJECT (mpeg2dec, "Too many decoding errors");
-          goto exit_error;
+        if (mpeg2dec->error_count >= WARN_THRESHOLD && WARN_THRESHOLD > 0) {
+          GST_ELEMENT_WARNING (mpeg2dec, STREAM, DECODE,
+              ("%d consecutive decoding errors", mpeg2dec->error_count),
+              (NULL));
         }
         goto exit;
       default:
@@ -1051,12 +1259,6 @@ exit:
     ret = GST_FLOW_OK;
     goto done;
   }
-exit_error:
-  {
-    GST_ELEMENT_ERROR (mpeg2dec, STREAM, DECODE, (NULL), (NULL));
-    ret = GST_FLOW_ERROR;
-    goto done;
-  }
 }
 
 static gboolean
@@ -1085,13 +1287,15 @@ gst_mpeg2dec_sink_event (GstPad * pad, GstEvent * event)
       if (format != GST_FORMAT_TIME)
         goto newseg_wrong_format;
 
-      if (rate <= 0.0)
-        goto newseg_wrong_rate;
-
       /* now configure the values */
       gst_segment_set_newsegment_full (&mpeg2dec->segment, update,
           rate, arate, format, start, stop, time);
 
+      GST_DEBUG_OBJECT (mpeg2dec,
+          "Pushing newseg rate %g, applied rate %g, format %d, start %"
+          G_GINT64_FORMAT ", stop %" G_GINT64_FORMAT ", pos %" G_GINT64_FORMAT,
+          rate, arate, format, start, stop, time);
+
       ret = gst_pad_push_event (mpeg2dec->srcpad, event);
       break;
     }
@@ -1105,13 +1309,16 @@ gst_mpeg2dec_sink_event (GstPad * pad, GstEvent * event)
       gst_mpeg2dec_qos_reset (mpeg2dec);
       mpeg2_reset (mpeg2dec->decoder, 0);
       mpeg2_skip (mpeg2dec->decoder, 1);
+      clear_queued (mpeg2dec);
       ret = gst_pad_push_event (mpeg2dec->srcpad, event);
       break;
     }
     case GST_EVENT_EOS:
+#ifndef GST_DISABLE_INDEX
       if (mpeg2dec->index && mpeg2dec->closed) {
         gst_index_commit (mpeg2dec->index, mpeg2dec->index_id);
       }
+#endif
       ret = gst_pad_push_event (mpeg2dec->srcpad, event);
       break;
     default:
@@ -1131,12 +1338,25 @@ newseg_wrong_format:
     gst_event_unref (event);
     goto done;
   }
-newseg_wrong_rate:
-  {
-    GST_DEBUG_OBJECT (mpeg2dec, "negative rates not supported yet");
-    gst_event_unref (event);
-    goto done;
-  }
+}
+
+static gboolean
+gst_mpeg2dec_setcaps (GstPad * pad, GstCaps * caps)
+{
+  GstMpeg2dec *mpeg2dec;
+  GstStructure *s;
+
+  mpeg2dec = GST_MPEG2DEC (gst_pad_get_parent (pad));
+
+  s = gst_caps_get_structure (caps, 0);
+
+  /* parse the par, this overrides the encoded par */
+  mpeg2dec->have_par = gst_structure_get_fraction (s, "pixel-aspect-ratio",
+      &mpeg2dec->pixel_width, &mpeg2dec->pixel_height);
+
+  gst_object_unref (mpeg2dec);
+
+  return TRUE;
 }
 
 static gboolean
@@ -1165,7 +1385,7 @@ gst_mpeg2dec_sink_convert (GstPad * pad, GstFormat src_format, gint64 src_value,
         case GST_FORMAT_TIME:
           if (info->sequence && info->sequence->byte_rate) {
             *dest_value =
-                gst_util_uint64_scale_int (GST_SECOND, src_value,
+                gst_util_uint64_scale (GST_SECOND, src_value,
                 info->sequence->byte_rate);
             GST_WARNING_OBJECT (mpeg2dec, "dest_value:%" GST_TIME_FORMAT,
                 GST_TIME_ARGS (*dest_value));
@@ -1380,6 +1600,7 @@ gst_mpeg2dec_get_event_masks (GstPad * pad)
 }
 #endif
 
+#ifndef GST_DISABLE_INDEX
 static gboolean
 index_seek (GstPad * pad, GstEvent * event)
 {
@@ -1446,7 +1667,7 @@ index_seek (GstPad * pad, GstEvent * event)
   }
   return FALSE;
 }
-
+#endif
 
 static gboolean
 normal_seek (GstPad * pad, GstEvent * event)
@@ -1532,9 +1753,11 @@ gst_mpeg2dec_src_event (GstPad * pad, GstEvent * event)
     case GST_EVENT_SEEK:{
       gst_event_ref (event);
       if (!(res = gst_pad_push_event (mpeg2dec->sinkpad, event))) {
+#ifndef GST_DISABLE_INDEX
         if (mpeg2dec->index)
           res = index_seek (pad, event);
         else
+#endif
           res = normal_seek (pad, event);
       }
       gst_event_unref (event);
@@ -1605,6 +1828,7 @@ gst_mpeg2dec_change_state (GstElement * element, GstStateChange transition)
       break;
     case GST_STATE_CHANGE_PAUSED_TO_READY:
       gst_mpeg2dec_qos_reset (mpeg2dec);
+      clear_queued (mpeg2dec);
       break;
     case GST_STATE_CHANGE_READY_TO_NULL:
       if (mpeg2dec->decoder) {
@@ -1628,40 +1852,10 @@ init_failed:
   }
 }
 
-static void
-gst_mpeg2dec_set_property (GObject * object, guint prop_id,
-    const GValue * value, GParamSpec * pspec)
-{
-  GstMpeg2dec *src;
-
-  g_return_if_fail (GST_IS_MPEG2DEC (object));
-  src = GST_MPEG2DEC (object);
-
-  switch (prop_id) {
-    default:
-      break;
-  }
-}
-
-static void
-gst_mpeg2dec_get_property (GObject * object, guint prop_id, GValue * value,
-    GParamSpec * pspec)
-{
-  GstMpeg2dec *mpeg2dec;
-
-  g_return_if_fail (GST_IS_MPEG2DEC (object));
-  mpeg2dec = GST_MPEG2DEC (object);
-
-  switch (prop_id) {
-    default:
-      break;
-  }
-}
-
 static gboolean
 plugin_init (GstPlugin * plugin)
 {
-  if (!gst_element_register (plugin, "mpeg2dec", GST_RANK_SECONDARY,
+  if (!gst_element_register (plugin, "mpeg2dec", GST_RANK_PRIMARY,
           GST_TYPE_MPEG2DEC))
     return FALSE;
 
