@@ -6,6 +6,7 @@
  * Copyright (C) <2005> Michael Smith <fluendo.com>
  * Copyright (C) <2006> Wim Taymans <wim@fluendo.com>
  * Copyright (C) <2006> Tim-Philipp Müller <tim centricular net>
+ * Copyright (C) <2007> Wim Taymans <wim.taymans@gmail.com>
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Library General Public
@@ -28,9 +29,7 @@
 #endif
 
 #include "rmdemux.h"
-#include "rdtdepay.h"
 #include "rmutils.h"
-#include "rademux.h"
 
 #include <string.h>
 #include <ctype.h>
@@ -41,6 +40,8 @@
 #define HEADER_SIZE 10
 #define DATA_SIZE 8
 
+#define MAX_FRAGS 256
+
 typedef struct _GstRMDemuxIndex GstRMDemuxIndex;
 
 struct _GstRMDemuxStream
@@ -53,6 +54,7 @@ struct _GstRMDemuxStream
   int id;
   GstPad *pad;
   GstFlowReturn last_flow;
+  gboolean discont;
   int timescale;
 
   int sample_index;
@@ -78,6 +80,18 @@ struct _GstRMDemuxStream
   guint subpackets_needed;      /* subpackets needed for descrambling    */
   GPtrArray *subpackets;        /* array containing subpacket GstBuffers */
 
+  /* Variables needed for fixing timestamps. */
+  GstClockTime next_ts, last_ts;
+  guint16 next_seq, last_seq;
+
+  gint frag_seqnum;
+  gint frag_subseq;
+  guint frag_length;
+  guint frag_current;
+  guint frag_count;
+  guint frag_offset[MAX_FRAGS];
+  GstAdapter *adapter;
+
   GstTagList *pending_tags;
 };
 
@@ -152,7 +166,7 @@ static void gst_rmdemux_parse_data (GstRMDemux * rmdemux, const guint8 * data,
 static void gst_rmdemux_parse_cont (GstRMDemux * rmdemux, const guint8 * data,
     int length);
 static GstFlowReturn gst_rmdemux_parse_packet (GstRMDemux * rmdemux,
-    const guint8 * data, guint16 version, guint16 length);
+    GstBuffer * in, guint16 version);
 static void gst_rmdemux_parse_indx_data (GstRMDemux * rmdemux,
     const guint8 * data, int length);
 static void gst_rmdemux_stream_clear_cached_subpackets (GstRMDemux * rmdemux,
@@ -247,6 +261,7 @@ gst_rmdemux_init (GstRMDemux * rmdemux)
 
   rmdemux->adapter = gst_adapter_new ();
   rmdemux->first_ts = GST_CLOCK_TIME_NONE;
+  rmdemux->base_ts = GST_CLOCK_TIME_NONE;
   rmdemux->need_newsegment = TRUE;
 }
 
@@ -445,6 +460,7 @@ find_seek_offset_time (GstRMDemux * rmdemux, GstClockTime time)
         break;
       }
     }
+    stream->discont = TRUE;
   }
   return ret;
 }
@@ -454,7 +470,7 @@ gst_rmdemux_perform_seek (GstRMDemux * rmdemux, GstEvent * event)
 {
   gboolean validated;
   gboolean ret = TRUE;
-  gboolean flush, accurate;
+  gboolean flush;
   GstFormat format;
   gdouble rate;
   GstSeekFlags flags;
@@ -489,7 +505,6 @@ gst_rmdemux_perform_seek (GstRMDemux * rmdemux, GstEvent * event)
   GST_DEBUG_OBJECT (rmdemux, "seek, rate %g", rate);
 
   flush = flags & GST_SEEK_FLAG_FLUSH;
-  accurate = flags & GST_SEEK_FLAG_ACCURATE;
 
   /* first step is to unlock the streaming thread if it is
    * blocked in a chain call, we do this by starting the flush. */
@@ -592,7 +607,7 @@ done:
   /* streaming can continue now */
   GST_PAD_STREAM_UNLOCK (rmdemux->sinkpad);
 
-  return TRUE;
+  return ret;
 
 error:
   {
@@ -679,6 +694,7 @@ gst_rmdemux_reset (GstRMDemux * rmdemux)
   for (cur = rmdemux->streams; cur; cur = cur->next) {
     GstRMDemuxStream *stream = cur->data;
 
+    g_object_unref (stream->adapter);
     gst_rmdemux_stream_clear_cached_subpackets (rmdemux, stream);
     gst_element_remove_pad (GST_ELEMENT (rmdemux), stream->pad);
     if (stream->pending_tags)
@@ -699,6 +715,7 @@ gst_rmdemux_reset (GstRMDemux * rmdemux)
 
   gst_segment_init (&rmdemux->segment, GST_FORMAT_UNDEFINED);
   rmdemux->first_ts = GST_CLOCK_TIME_NONE;
+  rmdemux->base_ts = GST_CLOCK_TIME_NONE;
   rmdemux->need_newsegment = TRUE;
 }
 
@@ -949,16 +966,25 @@ gst_rmdemux_chain (GstPad * pad, GstBuffer * buffer)
   GstFlowReturn ret = GST_FLOW_OK;
   const guint8 *data;
   guint16 version;
+  guint avail;
 
   GstRMDemux *rmdemux = GST_RMDEMUX (GST_PAD_PARENT (pad));
 
+  if (rmdemux->base_ts == -1) {
+    rmdemux->base_ts = GST_BUFFER_TIMESTAMP (buffer);
+    GST_LOG_OBJECT (rmdemux, "base_ts %" GST_TIME_FORMAT,
+        GST_TIME_ARGS (rmdemux->base_ts));
+  }
+
+  gst_adapter_push (rmdemux->adapter, buffer);
+
   GST_LOG_OBJECT (rmdemux, "Chaining buffer of size %d",
       GST_BUFFER_SIZE (buffer));
 
-  gst_adapter_push (rmdemux->adapter, buffer);
-
   while (TRUE) {
-    GST_LOG_OBJECT (rmdemux, "looping in chain");
+    avail = gst_adapter_available (rmdemux->adapter);
+
+    GST_LOG_OBJECT (rmdemux, "looping in chain, avail %u", avail);
     switch (rmdemux->state) {
       case RMDEMUX_STATE_HEADER:
       {
@@ -1153,20 +1179,30 @@ gst_rmdemux_chain (GstPad * pad, GstBuffer * buffer)
           data = gst_adapter_peek (rmdemux->adapter, 4);
 
           length = RMDEMUX_GUINT16_GET (data + 2);
+          GST_LOG_OBJECT (rmdemux, "Got length %d", length);
+
           if (length < 4) {
+            GST_LOG_OBJECT (rmdemux, "length too small, dropping");
             /* Invalid, just drop it */
             gst_adapter_flush (rmdemux->adapter, 4);
           } else {
-            if (gst_adapter_available (rmdemux->adapter) < length)
+            GstBuffer *buffer;
+
+            avail = gst_adapter_available (rmdemux->adapter);
+            if (avail < length)
               goto unlock;
-            data = gst_adapter_peek (rmdemux->adapter, length);
 
-            ret =
-                gst_rmdemux_parse_packet (rmdemux, data + 4, version,
-                length - 4);
+            GST_LOG_OBJECT (rmdemux, "we have %u available and we needed %d",
+                avail, length);
+
+            /* flush version and length */
+            gst_adapter_flush (rmdemux->adapter, 4);
+            length -= 4;
+
+            buffer = gst_adapter_take_buffer (rmdemux->adapter, length);
+
+            ret = gst_rmdemux_parse_packet (rmdemux, buffer, version);
             rmdemux->chunk_index++;
-
-            gst_adapter_flush (rmdemux->adapter, length);
           }
 
           if (rmdemux->chunk_index == rmdemux->n_chunks || length == 0)
@@ -1224,6 +1260,17 @@ gst_rmdemux_send_event (GstRMDemux * rmdemux, GstEvent * event)
     GST_DEBUG_OBJECT (rmdemux, "Pushing %s event on pad %s",
         GST_EVENT_TYPE_NAME (event), GST_PAD_NAME (stream->pad));
 
+    switch (GST_EVENT_TYPE (event)) {
+      case GST_EVENT_FLUSH_STOP:
+        stream->last_ts = -1;
+        stream->next_ts = -1;
+        stream->last_seq = -1;
+        stream->next_seq = -1;
+        stream->last_flow = GST_FLOW_OK;
+        break;
+      default:
+        break;
+    }
     gst_event_ref (event);
     gst_pad_push_event (stream->pad, event);
   }
@@ -1323,18 +1370,21 @@ gst_rmdemux_add_stream (GstRMDemux * rmdemux, GstRMDemuxStream * stream)
         stream->subpackets = NULL;
         break;
 
-        /* RealAudio 10 (AAC) */
-      case GST_RM_AUD_RAAC:
-        codec_name = "Real Audio 10 (AAC)";
-        version = 10;
-        break;
-
         /* MPEG-4 based */
+      case GST_RM_AUD_RAAC:
       case GST_RM_AUD_RACP:
-        /* FIXME: codec_name = */
+        codec_name = "MPEG4 audio";
         stream_caps =
             gst_caps_new_simple ("audio/mpeg", "mpegversion", G_TYPE_INT,
-            (int) 4, NULL);
+            (int) 4, "framed", G_TYPE_BOOLEAN, TRUE, NULL);
+        if (stream->extra_data_size > 0) {
+          /* strip off an unknown byte in the extra data */
+          stream->extra_data_size--;
+          stream->extra_data++;
+        }
+        stream->needs_descrambling = TRUE;
+        stream->subpackets_needed = 1;
+        stream->subpackets = NULL;
         break;
 
         /* Sony ATRAC3 */
@@ -1507,18 +1557,19 @@ gst_rmdemux_parse_mdpr (GstRMDemux * rmdemux, const guint8 * data, int length)
   int stream_type;
   int offset;
 
-  /* re_hexdump_bytes ((guint8 *) data, length, 0); */
-
   stream = g_new0 (GstRMDemuxStream, 1);
 
   stream->id = RMDEMUX_GUINT16_GET (data);
   stream->index = NULL;
   stream->seek_offset = 0;
+  stream->last_ts = -1;
+  stream->next_ts = -1;
   stream->last_flow = GST_FLOW_OK;
+  stream->discont = TRUE;
+  stream->adapter = gst_adapter_new ();
   GST_LOG_OBJECT (rmdemux, "stream_number=%d", stream->id);
 
   offset = 30;
-  stream_type = GST_RMDEMUX_STREAM_UNKNOWN;
   stream1_type_string = gst_rm_utils_read_string8 (data + offset,
       length - offset, &str_len);
   offset += str_len;
@@ -1577,8 +1628,8 @@ gst_rmdemux_parse_mdpr (GstRMDemux * rmdemux, const guint8 * data, int length)
       stream->rate = RMDEMUX_GUINT16_GET (data + offset + 16);
       stream->subformat = RMDEMUX_GUINT32_GET (data + offset + 26);
       stream->format = RMDEMUX_GUINT32_GET (data + offset + 30);
-      stream->extra_data_size = length - (offset + 34);
-      stream->extra_data = (guint8 *) data + offset + 34;
+      stream->extra_data_size = length - (offset + 26);
+      stream->extra_data = (guint8 *) data + offset + 26;
       /* Natural way to represent framerates here requires unsigned 32 bit
        * numerator, which we don't have. For the nasty case, approximate...
        */
@@ -1881,6 +1932,15 @@ gst_rmdemux_descramble_cook_audio (GstRMDemux * rmdemux,
     GstBuffer *subbuf;
 
     subbuf = gst_buffer_create_sub (outbuf, p * packet_size, packet_size);
+
+    GST_LOG_OBJECT (rmdemux, "pushing buffer timestamp %" GST_TIME_FORMAT,
+        GST_TIME_ARGS (GST_BUFFER_TIMESTAMP (subbuf)));
+
+    if (stream->discont) {
+      GST_BUFFER_FLAG_SET (subbuf, GST_BUFFER_FLAG_DISCONT);
+      stream->discont = FALSE;
+    }
+
     gst_buffer_set_caps (subbuf, GST_PAD_CAPS (stream->pad));
     ret = gst_pad_push (stream->pad, subbuf);
     if (ret != GST_FLOW_OK)
@@ -1907,9 +1967,58 @@ gst_rmdemux_descramble_dnet_audio (GstRMDemux * rmdemux,
   g_ptr_array_set_size (stream->subpackets, 0);
 
   buf = gst_rm_utils_descramble_dnet_buffer (buf);
+
+  if (stream->discont) {
+    GST_BUFFER_FLAG_SET (buf, GST_BUFFER_FLAG_DISCONT);
+    stream->discont = FALSE;
+  }
   return gst_pad_push (stream->pad, buf);
 }
 
+static GstFlowReturn
+gst_rmdemux_descramble_mp4a_audio (GstRMDemux * rmdemux,
+    GstRMDemuxStream * stream)
+{
+  GstFlowReturn res;
+  GstBuffer *buf, *outbuf;
+  guint frames, index, i;
+  guint8 *data;
+  GstClockTime timestamp;
+
+  res = GST_FLOW_OK;
+
+  buf = g_ptr_array_index (stream->subpackets, 0);
+  g_ptr_array_index (stream->subpackets, 0) = NULL;
+  g_ptr_array_set_size (stream->subpackets, 0);
+
+  data = GST_BUFFER_DATA (buf);
+  timestamp = GST_BUFFER_TIMESTAMP (buf);
+
+  frames = (data[1] & 0xf0) >> 4;
+  index = 2 * frames + 2;
+
+  for (i = 0; i < frames; i++) {
+    guint len = (data[i * 2 + 2] << 8) | data[i * 2 + 3];
+
+    outbuf = gst_buffer_create_sub (buf, index, len);
+    if (i == 0)
+      GST_BUFFER_TIMESTAMP (outbuf) = timestamp;
+    gst_buffer_set_caps (outbuf, GST_PAD_CAPS (stream->pad));
+
+    index += len;
+
+    if (stream->discont) {
+      GST_BUFFER_FLAG_SET (buf, GST_BUFFER_FLAG_DISCONT);
+      stream->discont = FALSE;
+    }
+    res = gst_pad_push (stream->pad, outbuf);
+    if (res != GST_FLOW_OK)
+      break;
+  }
+  gst_buffer_unref (buf);
+  return res;
+}
+
 static GstFlowReturn
 gst_rmdemux_handle_scrambled_packet (GstRMDemux * rmdemux,
     GstRMDemuxStream * stream, GstBuffer * buf, gboolean keyframe)
@@ -1940,6 +2049,10 @@ gst_rmdemux_handle_scrambled_packet (GstRMDemux * rmdemux,
     case GST_RM_AUD_COOK:
       ret = gst_rmdemux_descramble_cook_audio (rmdemux, stream);
       break;
+    case GST_RM_AUD_RAAC:
+    case GST_RM_AUD_RACP:
+      ret = gst_rmdemux_descramble_mp4a_audio (rmdemux, stream);
+      break;
     default:
       g_assert_not_reached ();
   }
@@ -1947,55 +2060,474 @@ gst_rmdemux_handle_scrambled_packet (GstRMDemux * rmdemux,
   return ret;
 }
 
+static GstClockTime
+gst_rmdemux_fix_timestamp (GstRMDemux * rmdemux, GstRMDemuxStream * stream,
+    guint8 * data, GstClockTime timestamp)
+{
+  guint8 frame_type;
+  guint16 seq;
+  GstClockTime ts = timestamp;
+
+  if (timestamp == GST_CLOCK_TIME_NONE)
+    goto done;
+
+  /* only adjust when we have a stream with B frames */
+  if (stream->format < 0x20200002)
+    goto done;
+
+  /* Fix timestamp. */
+  switch (stream->fourcc) {
+    case GST_RM_VDO_RV10:
+      goto done;
+    case GST_RM_VDO_RV20:
+    {
+      /*
+       * Bit  1- 2: frame type
+       * Bit  3- 9: ?
+       * Bit 10-22: sequence number
+       * Bit 23-32: ?
+       */
+      frame_type = (data[0] >> 6) & 0x03;
+      seq = ((data[1] & 0x7f) << 6) + ((data[2] & 0xfc) >> 2);
+      break;
+    }
+    case GST_RM_VDO_RV30:
+    {
+      /*
+       * Bit  1- 2: ?
+       * Bit     3: skip packet if 1
+       * Bit  4- 5: frame type
+       * Bit  6-12: ?
+       * Bit 13-25: sequence number
+       * Bit 26-32: ?
+       */
+      frame_type = (data[0] >> 3) & 0x03;
+      seq = ((data[1] & 0x0f) << 9) + (data[2] << 1) + ((data[3] & 0x80) >> 7);
+      break;
+    }
+    case GST_RM_VDO_RV40:
+    {
+      /*
+       * Bit     1: skip packet if 1
+       * Bit  2- 3: frame type
+       * Bit  4-13: ?
+       * Bit 14-26: sequence number
+       * Bit 27-32: ?
+       */
+      frame_type = (data[0] >> 5) & 0x03;
+      seq = ((data[1] & 0x07) << 10) + (data[2] << 2) + ((data[3] & 0xc0) >> 6);
+      break;
+    }
+    default:
+      goto unknown_version;
+  }
+
+  switch (frame_type) {
+    case 0:
+    case 1:
+    {
+      GST_LOG_OBJECT (rmdemux, "I frame %d", frame_type);
+      /* I frame */
+      if (stream->next_ts == -1)
+        stream->next_ts = timestamp;
+      else
+        timestamp = stream->next_ts;
+      stream->last_ts = stream->next_ts;
+      stream->next_ts = ts;
+      stream->last_seq = stream->next_seq;
+      stream->next_seq = seq;
+      break;
+    }
+    case 2:
+    {
+      GST_LOG_OBJECT (rmdemux, "P frame");
+      /* P frame */
+      timestamp = stream->last_ts = stream->next_ts;
+      if (seq < stream->next_seq)
+        stream->next_ts += (seq + 0x2000 - stream->next_seq) * GST_MSECOND;
+      else
+        stream->next_ts += (seq - stream->next_seq) * GST_MSECOND;
+      stream->last_seq = stream->next_seq;
+      stream->next_seq = seq;
+      break;
+    }
+    case 3:
+    {
+      GST_LOG_OBJECT (rmdemux, "B frame");
+      /* B frame */
+      if (seq < stream->last_seq) {
+        timestamp =
+            (seq + 0x2000 - stream->last_seq) * GST_MSECOND + stream->last_ts;
+      } else {
+        timestamp = (seq - stream->last_seq) * GST_MSECOND + stream->last_ts;
+      }
+      break;
+    }
+    default:
+      goto unknown_frame_type;
+  }
+
+done:
+  GST_LOG_OBJECT (rmdemux,
+      "timestamp %" GST_TIME_FORMAT " -> %" GST_TIME_FORMAT, GST_TIME_ARGS (ts),
+      GST_TIME_ARGS (timestamp));
+
+  return timestamp;
+
+  /* Errors */
+unknown_version:
+  {
+    GST_ELEMENT_ERROR (rmdemux, STREAM, DECODE,
+        ("Unknown version: %i.", stream->version), (NULL));
+    return GST_FLOW_ERROR;
+  }
+
+unknown_frame_type:
+  {
+    GST_ELEMENT_ERROR (rmdemux, STREAM, DECODE, ("Unknown frame type %d.",
+            frame_type), (NULL));
+    return GST_FLOW_ERROR;
+  }
+}
+
+#define PARSE_NUMBER(data, size, number, label) \
+G_STMT_START {                                  \
+  if (size < 2)                                 \
+    goto label;                                 \
+  number = GST_READ_UINT16_BE (data);           \
+  if (!(number & 0xc000)) {                     \
+    if (size < 4)                               \
+      goto label;                               \
+    number = GST_READ_UINT32_BE (data);         \
+    data += 4;                                  \
+    size -= 4;                                  \
+  } else {                                      \
+    number &= 0x3fff;                           \
+    data += 2;                                  \
+    size -= 2;                                  \
+  }                                             \
+} G_STMT_END
+
 static GstFlowReturn
-gst_rmdemux_parse_packet (GstRMDemux * rmdemux, const guint8 * data,
-    guint16 version, guint16 length)
+gst_rmdemux_parse_video_packet (GstRMDemux * rmdemux, GstRMDemuxStream * stream,
+    GstBuffer * in, guint offset, guint16 version,
+    GstClockTime timestamp, gboolean key)
+{
+  GstFlowReturn ret;
+  const guint8 *data, *base;
+  guint size;
+
+  base = GST_BUFFER_DATA (in);
+  data = base + offset;
+  size = GST_BUFFER_SIZE (in) - offset;
+  /* if size <= 2, we want this method to return the same GstFlowReturn as it
+   * was previously for that given stream. */
+  ret = stream->last_flow;
+
+  while (size > 2) {
+    guint8 pkg_header;
+    guint pkg_offset;
+    guint pkg_length;
+    guint pkg_subseq = 0, pkg_seqnum = G_MAXUINT;
+    guint fragment_size;
+    GstBuffer *fragment;
+
+    pkg_header = *data++;
+    size--;
+
+    /* packet header
+     * bit 7: 1=last block in block chain
+     * bit 6: 1=short header (only one block?)
+     */
+    if ((pkg_header & 0xc0) == 0x40) {
+      /* skip unknown byte */
+      data++;
+      size--;
+      pkg_offset = 0;
+      pkg_length = size;
+    } else {
+      if ((pkg_header & 0x40) == 0) {
+        pkg_subseq = (*data++) & 0x7f;
+        size--;
+      } else {
+        pkg_subseq = 0;
+      }
+
+      /* length */
+      PARSE_NUMBER (data, size, pkg_length, not_enough_data);
+
+      /* offset */
+      PARSE_NUMBER (data, size, pkg_offset, not_enough_data);
+
+      /* seqnum */
+      if (size < 1)
+        goto not_enough_data;
+
+      pkg_seqnum = *data++;
+      size--;
+    }
+
+    GST_DEBUG_OBJECT (rmdemux,
+        "seq %d, subseq %d, offset %d, length %d, size %d, header %02x",
+        pkg_seqnum, pkg_subseq, pkg_offset, pkg_length, size, pkg_header);
+
+    /* calc size of fragment */
+    if ((pkg_header & 0xc0) == 0x80) {
+      fragment_size = pkg_offset;
+    } else {
+      if ((pkg_header & 0xc0) == 0)
+        fragment_size = size;
+      else
+        fragment_size = pkg_length;
+    }
+    GST_DEBUG_OBJECT (rmdemux, "fragment size %d", fragment_size);
+
+    /* get the fragment */
+    fragment = gst_buffer_create_sub (in, data - base, fragment_size);
+
+    if (pkg_subseq == 1) {
+      GST_DEBUG_OBJECT (rmdemux, "start new fragment");
+      gst_adapter_clear (stream->adapter);
+      stream->frag_current = 0;
+      stream->frag_count = 0;
+      stream->frag_length = pkg_length;
+    } else if (pkg_subseq == 0) {
+      GST_DEBUG_OBJECT (rmdemux, "non fragmented packet");
+      stream->frag_current = 0;
+      stream->frag_count = 0;
+      stream->frag_length = fragment_size;
+    }
+
+    /* put fragment in adapter */
+    gst_adapter_push (stream->adapter, fragment);
+    stream->frag_offset[stream->frag_count] = stream->frag_current;
+    stream->frag_current += fragment_size;
+    stream->frag_count++;
+
+    if (stream->frag_count > MAX_FRAGS)
+      goto too_many_fragments;
+
+    GST_DEBUG_OBJECT (rmdemux, "stored fragment in adapter %d/%d",
+        stream->frag_current, stream->frag_length);
+
+    /* flush fragment when complete */
+    if (stream->frag_current >= stream->frag_length) {
+      GstBuffer *out;
+      guint8 *outdata;
+      guint header_size;
+      gint i, avail;
+
+      /* calculate header size, which is:
+       * 1 byte for the number of fragments - 1
+       * for each fragment:
+       *   4 bytes 0x00000001 little endian
+       *   4 bytes fragment offset
+       *
+       * This is also the matroska header for realvideo, the decoder needs the
+       * fragment offsets, both in ffmpeg and real .so, so we just give it that
+       * in front of the data.
+       */
+      header_size = 1 + (8 * (stream->frag_count));
+
+      GST_DEBUG_OBJECT (rmdemux,
+          "fragmented completed. count %d, header_size %u", stream->frag_count,
+          header_size);
+
+      avail = gst_adapter_available (stream->adapter);
+
+      out = gst_buffer_new_and_alloc (header_size + avail);
+      outdata = GST_BUFFER_DATA (out);
+
+      /* create header */
+      *outdata++ = stream->frag_count - 1;
+      for (i = 0; i < stream->frag_count; i++) {
+        GST_WRITE_UINT32_LE (outdata, 0x00000001);
+        outdata += 4;
+        GST_WRITE_UINT32_LE (outdata, stream->frag_offset[i]);
+        outdata += 4;
+      }
+
+      /* copy packet data after the header now */
+      gst_adapter_copy (stream->adapter, outdata, 0, avail);
+      gst_adapter_flush (stream->adapter, avail);
+
+      stream->frag_current = 0;
+      stream->frag_count = 0;
+      stream->frag_length = 0;
+
+      gst_buffer_set_caps (out, GST_PAD_CAPS (stream->pad));
+
+      if (timestamp != -1) {
+        if (rmdemux->first_ts != -1 && timestamp > rmdemux->first_ts)
+          timestamp -= rmdemux->first_ts;
+        else
+          timestamp = 0;
+
+        if (rmdemux->base_ts != -1)
+          timestamp += rmdemux->base_ts;
+      }
+      timestamp =
+          gst_rmdemux_fix_timestamp (rmdemux, stream, outdata, timestamp);
+
+      GST_BUFFER_TIMESTAMP (out) = timestamp;
+
+      GST_LOG_OBJECT (rmdemux, "pushing timestamp %" GST_TIME_FORMAT,
+          GST_TIME_ARGS (timestamp));
+
+      if (stream->discont) {
+        GST_BUFFER_FLAG_SET (out, GST_BUFFER_FLAG_DISCONT);
+        stream->discont = FALSE;
+      }
+
+      ret = gst_pad_push (stream->pad, out);
+      ret = gst_rmdemux_combine_flows (rmdemux, stream, ret);
+      if (ret != GST_FLOW_OK)
+        break;
+
+      timestamp = GST_CLOCK_TIME_NONE;
+    }
+    data += fragment_size;
+    size -= fragment_size;
+  }
+  GST_DEBUG_OBJECT (rmdemux, "%d bytes left", size);
+
+  gst_buffer_unref (in);
+
+  return ret;
+
+  /* ERRORS */
+not_enough_data:
+  {
+    GST_ELEMENT_WARNING (rmdemux, STREAM, DECODE, ("Skipping bad packet."),
+        (NULL));
+    gst_buffer_unref (in);
+    return GST_FLOW_OK;
+  }
+too_many_fragments:
+  {
+    GST_ELEMENT_ERROR (rmdemux, STREAM, DECODE,
+        ("Got more fragments (%u) than can be handled (%u)",
+            stream->frag_count, MAX_FRAGS), (NULL));
+    gst_buffer_unref (in);
+    return GST_FLOW_ERROR;
+  }
+}
+
+static GstFlowReturn
+gst_rmdemux_parse_audio_packet (GstRMDemux * rmdemux, GstRMDemuxStream * stream,
+    GstBuffer * in, guint offset, guint16 version,
+    GstClockTime timestamp, gboolean key)
+{
+  GstFlowReturn ret, cret;
+  GstBuffer *buffer;
+  const guint8 *data;
+  guint size;
+
+  data = GST_BUFFER_DATA (in) + offset;
+  size = GST_BUFFER_SIZE (in) - offset;
+
+  ret = gst_pad_alloc_buffer_and_set_caps (stream->pad,
+      GST_BUFFER_OFFSET_NONE, size, GST_PAD_CAPS (stream->pad), &buffer);
+
+  cret = gst_rmdemux_combine_flows (rmdemux, stream, ret);
+  if (ret != GST_FLOW_OK)
+    goto alloc_failed;
+
+  memcpy (GST_BUFFER_DATA (buffer), (guint8 *) data, size);
+
+  if (rmdemux->first_ts != -1 && timestamp > rmdemux->first_ts)
+    timestamp -= rmdemux->first_ts;
+  else
+    timestamp = 0;
+
+  if (rmdemux->base_ts != -1)
+    timestamp += rmdemux->base_ts;
+
+  GST_BUFFER_TIMESTAMP (buffer) = timestamp;
+
+  if (stream->needs_descrambling) {
+    GST_LOG_OBJECT (rmdemux, "descramble timestamp %" GST_TIME_FORMAT,
+        GST_TIME_ARGS (timestamp));
+    ret = gst_rmdemux_handle_scrambled_packet (rmdemux, stream, buffer, key);
+  } else {
+    GST_LOG_OBJECT (rmdemux,
+        "Pushing buffer of size %d, timestamp %" GST_TIME_FORMAT "to pad %s",
+        GST_BUFFER_SIZE (buffer), GST_TIME_ARGS (timestamp),
+        GST_PAD_NAME (stream->pad));
+
+    if (stream->discont) {
+      GST_BUFFER_FLAG_SET (buffer, GST_BUFFER_FLAG_DISCONT);
+      stream->discont = FALSE;
+    }
+    ret = gst_pad_push (stream->pad, buffer);
+  }
+  return ret;
+
+  /* ERRORS */
+alloc_failed:
+  {
+    GST_DEBUG_OBJECT (rmdemux, "pad alloc returned %d", ret);
+    return cret;
+  }
+}
+
+static GstFlowReturn
+gst_rmdemux_parse_packet (GstRMDemux * rmdemux, GstBuffer * in, guint16 version)
 {
   guint16 id;
   GstRMDemuxStream *stream;
-  GstBuffer *buffer;
-  guint16 packet_size;
+  guint size;
   GstFlowReturn cret, ret;
   GstClockTime timestamp;
-  gboolean key = FALSE;
+  gboolean key;
+  guint8 *data, *base;
+  guint8 flags;
+  guint32 ts;
 
+  base = data = GST_BUFFER_DATA (in);
+  size = GST_BUFFER_SIZE (in);
+
+  /* stream number */
   id = RMDEMUX_GUINT16_GET (data);
-  /* timestamp in Msec */
-  timestamp = RMDEMUX_GUINT32_GET (data + 2) * GST_MSECOND;
-
-  gst_segment_set_last_stop (&rmdemux->segment, GST_FORMAT_TIME, timestamp);
-
-  GST_LOG_OBJECT (rmdemux, "Parsing a packet for stream=%d, timestamp=%"
-      GST_TIME_FORMAT ", version=%d", id, GST_TIME_ARGS (timestamp), version);
-
-  /* skip stream_id and timestamp */
-  data += 2 + 4;
-  packet_size = length - (2 + 4);
-
-  /* skip other stuff */
-  if (version == 0) {
-    /* uint8 packet_group */
-    /* uint8 flags        */
-    key = ((GST_READ_UINT8 (data + 1) & 0x02) != 0);
-    data += 2;
-    packet_size -= 2;
-  } else {
-    /* uint16 asm_rule */
-    /* uint8 asm_flags */
-    data += 3;
-    packet_size -= 3;
-  }
 
   stream = gst_rmdemux_get_stream_by_id (rmdemux, id);
   if (!stream || !stream->pad)
     goto unknown_stream;
 
+  /* timestamp in Msec */
+  ts = RMDEMUX_GUINT32_GET (data + 2);
+  timestamp = ts * GST_MSECOND;
+
+  gst_segment_set_last_stop (&rmdemux->segment, GST_FORMAT_TIME, timestamp);
+
+  GST_LOG_OBJECT (rmdemux, "Parsing a packet for stream=%d, timestamp=%"
+      GST_TIME_FORMAT ", size %u, version=%d, ts=%u", id,
+      GST_TIME_ARGS (timestamp), size, version, ts);
+
   if (rmdemux->first_ts == GST_CLOCK_TIME_NONE) {
     GST_DEBUG_OBJECT (rmdemux, "First timestamp: %" GST_TIME_FORMAT,
         GST_TIME_ARGS (timestamp));
     rmdemux->first_ts = timestamp;
   }
 
+  /* skip stream_id and timestamp */
+  data += (2 + 4);
+  size -= (2 + 4);
+
+  /* get flags */
+  flags = GST_READ_UINT8 (data + 1);
+
+  data += 2;
+  size -= 2;
+
+  /* version 1 has an extra byte */
+  if (version == 1) {
+    data += 1;
+    size -= 1;
+  }
+  key = (flags & 0x02) != 0;
+  GST_DEBUG_OBJECT (rmdemux, "flags %d, Keyframe %d", flags, key);
+
   if (rmdemux->need_newsegment) {
     GstEvent *event;
 
@@ -2017,32 +2549,25 @@ gst_rmdemux_parse_packet (GstRMDemux * rmdemux, const guint8 * data,
     stream->pending_tags = NULL;
   }
 
-  if ((rmdemux->offset + packet_size) <= stream->seek_offset) {
+  if ((rmdemux->offset + size) <= stream->seek_offset) {
     GST_DEBUG_OBJECT (rmdemux,
-        "Stream %d is skipping: seek_offset=%d, offset=%d, packet_size=%u",
-        stream->id, stream->seek_offset, rmdemux->offset, packet_size);
+        "Stream %d is skipping: seek_offset=%d, offset=%d, size=%u",
+        stream->id, stream->seek_offset, rmdemux->offset, size);
     cret = GST_FLOW_OK;
     goto beach;
   }
 
-  ret = gst_pad_alloc_buffer_and_set_caps (stream->pad,
-      GST_BUFFER_OFFSET_NONE, packet_size, GST_PAD_CAPS (stream->pad), &buffer);
-
-  cret = gst_rmdemux_combine_flows (rmdemux, stream, ret);
-  if (ret != GST_FLOW_OK)
-    goto alloc_failed;
-
-  memcpy (GST_BUFFER_DATA (buffer), (guint8 *) data, packet_size);
-  GST_BUFFER_TIMESTAMP (buffer) = timestamp - rmdemux->first_ts;
-
-  if (stream->needs_descrambling) {
-    ret = gst_rmdemux_handle_scrambled_packet (rmdemux, stream, buffer, key);
-  } else {
-    GST_LOG_OBJECT (rmdemux, "Pushing buffer of size %d to pad %s",
-        GST_BUFFER_SIZE (buffer), GST_PAD_NAME (stream->pad));
-
-    ret = gst_pad_push (stream->pad, buffer);
-  }
+  /* do special headers */
+  if (stream->subtype == GST_RMDEMUX_STREAM_VIDEO) {
+    ret =
+        gst_rmdemux_parse_video_packet (rmdemux, stream, in, data - base,
+        version, timestamp, key);
+  } else if (stream->subtype == GST_RMDEMUX_STREAM_AUDIO) {
+    ret =
+        gst_rmdemux_parse_audio_packet (rmdemux, stream, in, data - base,
+        version, timestamp, key);
+  } else
+    ret = GST_FLOW_OK;
 
   cret = gst_rmdemux_combine_flows (rmdemux, stream, ret);
 
@@ -2056,37 +2581,11 @@ unknown_stream:
         "data packet", id);
     return GST_FLOW_OK;
   }
-alloc_failed:
-  {
-    GST_DEBUG_OBJECT (rmdemux, "pad alloc returned %d", ret);
-    return cret;
-  }
 }
 
-static gboolean
+gboolean
 gst_rmdemux_plugin_init (GstPlugin * plugin)
 {
   return gst_element_register (plugin, "rmdemux",
-      GST_RANK_PRIMARY, GST_TYPE_RMDEMUX) &&
-      gst_element_register (plugin, "rademux",
-      GST_RANK_SECONDARY, GST_TYPE_REAL_AUDIO_DEMUX);
+      GST_RANK_PRIMARY, GST_TYPE_RMDEMUX);
 }
-
-
-static gboolean
-plugin_init (GstPlugin * plugin)
-{
-  if (!gst_rmdemux_plugin_init (plugin))
-    return FALSE;
-
-  if (!gst_rdt_depay_plugin_init (plugin))
-    return FALSE;
-
-  return TRUE;
-}
-
-GST_PLUGIN_DEFINE (GST_VERSION_MAJOR,
-    GST_VERSION_MINOR,
-    "realmedia",
-    "RealMedia demuxer and depayloader",
-    plugin_init, VERSION, "LGPL", GST_PACKAGE_NAME, GST_PACKAGE_ORIGIN);
